# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Модуль фильтров языковой модели по разным признакам.

import numpy as np
from typing import List
from collections import defaultdict

from rupo.rhymes.rhymes import Rhymes
from rupo.main.markup import Word
from rupo.util.vocabulary import Vocabulary


class Filter(object):
    """
    Фильтр языковой модели.
    """
    def filter_word(self, word: Word) -> bool:
        raise NotImplementedError()

    def pass_word(self, word: Word) -> None:
        raise NotImplementedError()

    def filter_model(self, model: np.array, vocabulary: Vocabulary) -> np.array:
        """
        Фильтрация языковой модели.

        :param model: изначальная модель.
        :param vocabulary: словарь
        :return: модель после фильтрации и нормирования.
        """
        for i in range(len(model)):
            if not self.filter_word(vocabulary.get_word(i)):
                model[i] = 0.0
        s = sum(model)
        for i in range(len(model)):
            model[i] = model[i]/s if s != 0.0 else 0.0
        return model

    def filter_words(self, words: List[Word]) -> List[Word]:
        """
        Фильтрация набора слов.

        :param words: слова.
        :return: слова после фильтрации.
        """
        return [word for word in words if self.filter_word(word)]


class MetreFilter(Filter):
    """
    Фильтр по шаблону метра.
    """
    def __init__(self, metre_pattern: str):
        self.metre_pattern = metre_pattern
        self.position = len(metre_pattern) - 1

    def filter_word(self, word: Word) -> bool:
        """
        Фильтрация слова по метру в текущей позиции.

        :param word: слово.
        :return: подходит слово или нет.
        """
        syllables_count = len(word.syllables)
        if syllables_count > self.position + 1:
            return False
        for i in range(syllables_count):
            syllable = word.syllables[i]
            syllable_number = self.position - syllables_count + i + 1
            if syllables_count >= 2 and syllable.accent == -1 and self.metre_pattern[syllable_number] == "+":
                for j in range(syllables_count):
                    other_syllable = word.syllables[j]
                    other_syllable_number = other_syllable.number - syllable.number + syllable_number
                    if i != j and other_syllable.accent != -1 and self.metre_pattern[other_syllable_number] == "-":
                        return False
        return True

    def pass_word(self, word: Word) -> None:
        """
        Сдвинуть позицию в шаблоне метра на слово.

        :param word: слово (Word).
        """
        self.position -= len(word.syllables)

    def reset(self) -> None:
        """
        Сброс позиции в шаблоне.
        """
        self.position = len(self.metre_pattern) - 1


class RhymeFilter(Filter):
    """
    Фильтр по шаблону рифмы.
    """
    def __init__(self, rhyme_pattern: str, letters_to_rhymes: dict=None):
        self.rhyme_pattern = rhyme_pattern
        self.position = len(self.rhyme_pattern) - 1
        self.letters_to_rhymes = defaultdict(set)
        if letters_to_rhymes is not None:
            for letter, words in letters_to_rhymes.items():
                for word in words:
                    self.letters_to_rhymes[letter].add(word)

    def filter_word(self, word: Word) -> bool:
        """
        Фильтрация слова по рифме в текущей позиции.

        :param word: слово (Word).
        :return: подходит слово или нет.
        """
        if len(word.syllables) <= 1:
            return False
        if len(self.letters_to_rhymes[self.rhyme_pattern[self.position]]) == 0:
            return True
        first_word = list(self.letters_to_rhymes[self.rhyme_pattern[self.position]])[0]

        is_rhyme = Rhymes.is_rhyme(first_word, word, score_border=5, syllable_number_border=2) and \
            first_word.text != word.text
        return is_rhyme

    def pass_word(self, word: Word) -> None:
        """
        Сдвинуть позицию в шаблоне рифмы на строчку.

        :param word: рифмующееся слово (Word).
        """
        self.letters_to_rhymes[self.rhyme_pattern[self.position]].add(word)
        self.position -= 1


# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Индексы слов для языковой модели.

from typing import Dict, List, Set

from rupo.main.markup import Word


class Vocabulary(object):
    """
    Индексированный словарь.
    """
    def __init__(self) -> None:
        self.word_to_index = {}  # type: Dict[str, int]
        self.words = []  # type: List[Word]
        self.shorts_set = set()  # type: Set[str]

    def add_word(self, word: Word) -> bool:
        """
        Добавление слова.

        :param word: слово.
        :return: слово новое или нет.
        """
        short = word.get_short()
        if short not in self.shorts_set:
            self.words.append(word)
            self.shorts_set.add(short)
            self.word_to_index[short] = len(self.words)-1
            return True
        return False

    def get_word_index(self, word: Word) -> int:
        """
        Получить индекс слова.

        :param word: слово (Word).
        :return: индекс.
        """
        short = word.get_short()
        if short in self.word_to_index:
            return self.word_to_index[short]
        raise IndexError("Can't find word: " + word.text)

    def get_word(self, index: int) -> Word:
        """
        Получить слово по индексу.

        :param index: индекс.
        :return: слово.
        """
        return self.words[index]

    def shrink(self, short_words: List[str]) -> None:
        """
        Обрезать словарь по заданным коротким формам слов.

        :param short_words: короткие формы слов.
        """
        new_words = []
        new_shorts_set = set()
        short_words = set(short_words)
        for word in self.words:
            short = word.get_short()
            if short in short_words:
                new_words.append(word)
                new_shorts_set.add(short)
                self.word_to_index[short] = len(new_words)-1
        self.shorts_set = new_shorts_set
        self.words = new_words




# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Тесты к компилятору выражений.

import unittest

from rupo.metre.patterns import Patterns


class TestPatterns(unittest.TestCase):
    def test_patterns(self):
        self.assertCountEqual(Patterns.compile_pattern("(s)*", 3), ["sss"])
        self.assertCountEqual(Patterns.compile_pattern("(s)?(u)?(S)?", 2), ["su", "uS", "sS"])
        self.assertCountEqual(Patterns.compile_pattern("(s)?(u)(s)*", 1), ["u"])
        self.assertCountEqual(Patterns.compile_pattern("(s)?(u)(s)*", 2), ["su", "us"])
        self.assertCountEqual(Patterns.compile_pattern("(s)?(u)(s)*", 3), ["sus", "uss"])
        self.assertCountEqual(Patterns.compile_pattern("(us)*(uS)(U)?(U)?", 4), ["usuS", "uSUU"])
        self.assertCountEqual(Patterns.compile_pattern("((s)(u)?)*", 4), ["susu", "ssss", "suss", 'sssu', 'ssus'])





# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Классификатор метра.

from collections import OrderedDict
from typing import List, Dict, Tuple
import jsonpickle

from rupo.main.markup import Line, Markup
from rupo.util.mixins import CommonMixin
from rupo.metre.patterns import CompiledPatterns
from rupo.util.preprocess import get_first_vowel_position
from rupo.accents.classifier import MLAccentClassifier


class AccentCorrection(CommonMixin):
    """
    Исправление ударения.
    """
    def __init__(self, line_number: int, word_number: int, syllable_number: int,
                 word_text: str, accent: int) -> None:
        """
        :param line_number: номер строки.
        :param word_number: номер слова.
        :param syllable_number: номер слога.
        :param word_text: текст слова.
        :param accent: позиция ударения (с 0).
        """
        self.line_number = line_number
        self.word_number = word_number
        self.syllable_number = syllable_number
        self.word_text = word_text
        self.accent = accent


class LineClassificationResult(CommonMixin):
    """
    Результат классификации строки.
    """
    def __init__(self) -> None:
        self.error_count = {metre_name: 100 for metre_name in MetreClassifier.metres.keys()}
        self.best_patterns = {metre_name: "" for metre_name in MetreClassifier.metres.keys()}

    def store_best_result(self, metre_name: str, pattern: str, error_count: int) -> None:
        """
        Сохранить результат сравнения строки с шаблоном метра.

        :param metre_name: имя метра.
        :param pattern: шаблон метра.
        :param error_count: количество ошибок.
        """
        if error_count < self.error_count[metre_name]:
            self.error_count[metre_name] = error_count
            self.best_patterns[metre_name] = pattern

    def get_best_pattern(self, metre_name: str) -> str:
        """
        :return: лучший шаблон метра.
        """
        return self.best_patterns[metre_name]

    def get_best_metres(self) -> List[str]:
        """
        :return: метры с наименьшим количеством ошибок.
        """
        line_metres = []
        for key in self.error_count:
            self.error_count[key] *= MetreClassifier.line_inner_coef[key]
        min_errors = self.error_count[min(self.error_count, key=self.error_count.get)]
        for metre_name, error_count in self.error_count.items():
            if error_count == min_errors:
                line_metres.append(metre_name)
        return line_metres


class ClassificationResult(CommonMixin):
    """
    Результат классификации стихотворения по метру.
    """
    def __init__(self, count_lines: int=0) -> None:
        """
        :param count_lines: количество строк.
        """
        self.metre = None
        self.count_lines = count_lines
        self.lines_result = [LineClassificationResult() for i in range(count_lines)]
        self.errors_count = {k: 0 for k in MetreClassifier.metres.keys()}  # type: Dict[str, int]
        self.corrections = {k: [] for k in MetreClassifier.metres.keys()}  # type: Dict[str, List[AccentCorrection]]
        self.resolutions = {k: [] for k in MetreClassifier.metres.keys()}  # type: Dict[str, List[AccentCorrection]]
        self.additions = {k: [] for k in MetreClassifier.metres.keys()}  # type: Dict[str, List[AccentCorrection]]
        self.ml_resolutions = []  # type: List[AccentCorrection]

    def get_metre_errors_count(self):
        """
        :return: получить количество ошибок на заданном метре.
        """
        return self.errors_count[self.metre]

    def to_json(self):
        """
        :return: сериализация в json.
        """
        return jsonpickle.encode(self)

    @staticmethod
    def str_corrections(collection: List[AccentCorrection]) -> str:
        """
        :param collection: список исправлений.
        :return: его строковое представление.
        """
        return"\n".join([str((item.word_text, item.syllable_number)) for item in collection])

    def __str__(self):
        st = "Метр: " + str(self.metre) + "\n"
        st += "Снятая омография: \n" + ClassificationResult.str_corrections(self.resolutions[self.metre]) + "\n"
        st += "Неправильные ударения: \n" + ClassificationResult.str_corrections(self.corrections[self.metre]) + "\n"
        st += "Новые ударения: \n" + ClassificationResult.str_corrections(self.additions[self.metre]) + "\n"
        st += "ML: \n" + ClassificationResult.str_corrections(self.ml_resolutions) + "\n"
        return st


class CompilationsSingleton:
    compilations = None

    @classmethod
    def get(cls):
        if cls.compilations is None:
            cls.compilations = CompiledPatterns(MetreClassifier.metres, MetreClassifier.border_syllables_count)
        return cls.compilations


class MetreClassifier(object):
    """
    Классификатор, считает отклонения от стандартных шаблонов ритма(метров).
    """
    metres = OrderedDict(
        [("iambos", '(us)*(us)(u)?(u)?'),
         ("choreios", '(su)*(s)(u)?(u)?'),
         ("daktylos", '(suu)*(s)(u)?(u)?'),
         ("amphibrachys", '(usu)*(us)(u)?(u)?'),
         ("anapaistos",  '(uus)*(uus)(u)?(u)?'),
         ("dolnik3", '(u)?(u)?((su)(u)?)*(s)(u)?(u)?'),
         ("dolnik2", '(u)?(u)?((s)(u)?)*(s)(u)?(u)?'),
         ("taktovik3", '(u)?(u)?((su)(u)?(u)?)*(s)(u)?(u)?'),
         ("taktovik2", '(u)?(u)?((s)(u)?(u)?)*(s)(u)?(u)?')
         ])

    line_inner_coef = OrderedDict(
        [("iambos", 1.0),
         ("choreios", 1.0),
         ("daktylos", 1.2),
         ("amphibrachys", 1.2),
         ("anapaistos",  1.2),
         ("dolnik3", 2.0),
         ("dolnik2", 2.0),
         ("taktovik3", 3.0),
         ("taktovik2", 3.0)
         ])

    line_outer_coef = OrderedDict(
        [("iambos", 1.0),
         ("choreios", 1.0),
         ("daktylos", 1.0),
         ("amphibrachys", 1.0),
         ("anapaistos",  1.0),
         ("dolnik3", 0.8),
         ("dolnik2", 0.8),
         ("taktovik3", 0.6),
         ("taktovik2", 0.6)
         ])

    border_syllables_count = 18

    @staticmethod
    def classify_metre(markup):
        """
        Классифицируем стихотворный метр.

        :param markup: разметка.
        :return: результат классификации.
        """
        result = ClassificationResult(len(markup.lines))
        line_clf_results = [LineClassificationResult() for i in range(len(markup.lines))]
        for l in range(len(markup.lines)):
            for metre_name, metre_pattern in MetreClassifier.metres.items():
                line = markup.lines[l]
                line_syllables_count = sum([len(word.syllables) for word in line.words])

                # Строчки длиной больше border_syllables_count слогов не обрабатываем.
                if line_syllables_count > MetreClassifier.border_syllables_count or line_syllables_count == 0:
                    continue
                # Используем запомненные шаблоны.
                patterns = CompilationsSingleton.get().get_patterns(metre_name, line_syllables_count)

                # Сохраняем результаты по всем шаблонам.
                if len(patterns) == 0:
                    continue
                for pattern in patterns:
                    if pattern == "":
                        continue
                    error_count = MetreClassifier.__line_pattern_matching(line, pattern)
                    line_clf_results[l].store_best_result(metre_name, pattern, error_count)

        lines_metres = [line_clf_results[l].get_best_metres() for l in range(len(markup.lines))]
        # Выбираем общий метр по метрам строк с учётом коэффициентов.
        counter = {k: 0 for k in MetreClassifier.metres.keys()}
        for l in range(len(markup.lines)):
            for metre in lines_metres[l]:
                counter[metre] += 1
        for key in counter.keys():
            counter[key] *= MetreClassifier.line_outer_coef[key]
        result.metre = max(counter, key=counter.get)

        # Запомним все исправления.
        for l in range(len(markup.lines)):
            pattern = line_clf_results[l].get_best_pattern(result.metre)
            if pattern == "":
                continue
            corrections, resolutions, additions =\
                MetreClassifier.__get_line_pattern_matching_corrections(markup.lines[l], l, pattern)
            result.corrections[result.metre] += corrections
            result.resolutions[result.metre] += resolutions
            result.additions[result.metre] += additions
            result.errors_count[result.metre] += len(corrections)
        return result

    @staticmethod
    def __line_pattern_matching(line: Line, pattern: str) -> int:
        """
        Сопоставляем строку шаблону, считаем ошибки.

        :param line: строка.
        :param pattern: шаблон.
        :return: количество ошибок
        """
        error_count = 0
        number_in_pattern = 0
        for w in range(len(line.words)):
            word = line.words[w]
            # Игнорируем слова длиной меньше 2 слогов.
            if len(word.syllables) > 1:
                numbers = word.get_accented_syllables_numbers()
                is_error = sum(1 for number in numbers if pattern[number_in_pattern+number] == "s") == 0
                if is_error:
                    error_count += 1
            number_in_pattern += len(word.syllables)
        return error_count

    @staticmethod
    def __get_line_pattern_matching_corrections(line: Line, line_number: int, pattern: str) \
            -> Tuple[List[AccentCorrection], List[AccentCorrection], List[AccentCorrection]]:
        """
        Ударения могут приходиться на слабое место,
        если безударный слог того же слова не попадает на икт. Иначе - ошибка.

        :param line: строка.
        :param line_number: номер строки.
        :param pattern: шаблон.
        :return: ошибки, дополнения и снятия
        """
        corrections = []
        resolutions = []
        additions = []
        number_in_pattern = 0
        for w in range(len(line.words)):
            word = line.words[w]
            # Игнорируем слова длиной меньше 2 слогов.
            if len(word.syllables) <= 1:
                number_in_pattern += len(word.syllables)
                continue
            accents_count = word.count_accents()
            for syllable in word.syllables:
                if accents_count == 0 and pattern[number_in_pattern] == "s":
                    # Ударений нет, ставим такое, какое подходит по метру. Возможно несколько.
                    additions.append(AccentCorrection(line_number, w, syllable.number, word.text, syllable.vowel()))
                elif pattern[number_in_pattern] == "u" and syllable.accent != -1:
                    # Ударение есть и оно падает на этот слог, при этом в шаблоне безударная позиция.
                    # Найдём такой слог, у которого в шаблоне ударная позиция. Это и есть наше исправление.
                    for other_syllable in word.syllables:
                        other_number_in_pattern = other_syllable.number - syllable.number + number_in_pattern
                        if syllable.number == other_syllable.number or pattern[other_number_in_pattern] != "s":
                            continue
                        ac = AccentCorrection(line_number, w, other_syllable.number, word.text, other_syllable.vowel())
                        if accents_count == 1 and other_syllable.accent == -1:
                            corrections.append(ac)
                        else:
                            resolutions.append(ac)
                number_in_pattern += 1
        return corrections, resolutions, additions

    @staticmethod
    def get_ml_resolutions(result: ClassificationResult, accent_classifier: MLAccentClassifier) -> None:
        """
        Пытаемся снять омонимию с предыдущих этапов древесным классификатором.

        :param result: результат классификации.
        :param accent_classifier: древесный классификатор ударений.
        """
        result_additions = result.additions[result.metre]  # type: List[AccentCorrection]
        for i in range(len(result_additions)):
            for j in range(i, len(result_additions)):
                text1 = result_additions[i].word_text
                text2 = result_additions[j].word_text
                number1 = result_additions[i].syllable_number
                number2 = result_additions[j].syllable_number
                if text1 == text2 and number1 != number2:
                    accent = accent_classifier.classify_accent(text1)
                    if accent == number1:
                        result.ml_resolutions.append(result_additions[i])
                    if accent == number2:
                        result.ml_resolutions.append(result_additions[j])

    @staticmethod
    def get_improved_markup(markup: Markup, result: ClassificationResult) -> Markup:
        """
        Улучшаем разметку после классификации метра.

        :param markup: начальная разметка.
        :param result: результат классификации.
        :return: улучшенная разметка.
        """
        for pos in result.corrections[result.metre] + result.resolutions[result.metre]:
            syllables = markup.lines[pos.line_number].words[pos.word_number].syllables
            for i in range(len(syllables)):
                syllable = syllables[i]
                syllable.accent = -1
                if syllable.number == pos.syllable_number:
                    syllable.accent = syllable.begin + get_first_vowel_position(syllable.text)

        for pos in result.additions[result.metre]:
            syllable = markup.lines[pos.line_number].words[pos.word_number].syllables[pos.syllable_number]
            syllable.accent = syllable.begin + get_first_vowel_position(syllable.text)

        for pos in result.ml_resolutions:
            syllables = markup.lines[pos.line_number].words[pos.word_number].syllables
            for i in range(len(syllables)):
                syllable = syllables[i]
                syllable.accent = -1
                if syllable.number == pos.syllable_number:
                    syllable.accent = syllable.begin + get_first_vowel_position(syllable.text)
        return markup

    @staticmethod
    def improve_markup(markup: Markup, accents_classifier: MLAccentClassifier=None) -> \
            Tuple[Markup, ClassificationResult]:
        """
        Улучшение разметки метрическим и машинным классификатором.

        :param markup: начальная разметка.
        :param accents_classifier: классификатор ударений.
        """
        result = MetreClassifier.classify_metre(markup)
        if accents_classifier is not None:
            MetreClassifier.get_ml_resolutions(result, accents_classifier)
        return MetreClassifier.get_improved_markup(markup, result), result



from pkg_resources import resource_filename
foo_config = resource_filename(__name__, 'foo.conf')

DICT_TXT_PATH = resource_filename(__name__, "data/dict/accents_dict.txt")
DICT_TRIE_PATH = resource_filename(__name__, "data/dict/accents_dict.trie")
CLASSIFIER_PATH = resource_filename(__name__, "data/classifier/")
MARKUP_XML_EXAMPLE = resource_filename(__name__, "data/examples/markup.xml")
MARKUP_JSON_EXAMPLE = resource_filename(__name__, "data/examples/markup.json")
TEXT_XML_EXAMPLE = resource_filename(__name__, "data/examples/text.xml")
TEXT_TXT_EXAMPLE = resource_filename(__name__, "data/examples/text.txt")
EXAMPLES_DIR = resource_filename(__name__, "data/examples/")


# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Расстановка ударений на основе машинного обучения.

import os
from typing import List, Tuple

import numpy as np
from sklearn.externals import joblib
from sklearn.model_selection import ShuffleSplit, cross_val_score
from sklearn.tree import DecisionTreeClassifier

from rupo.accents.dict import AccentDict
from rupo.main.markup import Syllable
from rupo.main.phonetics import Phonetics
from rupo.util.preprocess import CYRRILIC_LOWER_CONSONANTS, CYRRILIC_LOWER_VOWELS, VOWELS
from rupo.util.preprocess import get_first_vowel_position
from rupo.settings import CLASSIFIER_PATH


class MLAccentClassifier:
    """
    Классификатор ударений на основе машинного обучения.
    """

    clf_filename = "clf_{}.pickle"

    def __init__(self, accents_dict: AccentDict) -> None:
        """
        :param accents_dict: словарь ударений.
        """
        if not os.path.exists(CLASSIFIER_PATH):
            os.mkdir(CLASSIFIER_PATH)
        for l in range(2, 13):
            if not os.path.isfile(os.path.join(CLASSIFIER_PATH, self.clf_filename.format(l))):
                self.__build_accent_classifier(CLASSIFIER_PATH, accents_dict, l)
        self.classifiers = {l: joblib.load(os.path.join(CLASSIFIER_PATH,
                                                        self.clf_filename.format(l))) for l in range(2, 13)}

    def do_cross_val(self, accents_dict: AccentDict) -> List[float]:
        """
        Кроссвалидация классификаторов.

        :param accents_dict: словарь ударений.
        :return result: среднее по кроссвалидации каждого классификатора.
        """
        result = []
        for l in range(2, 13):
            train_data, answers = MLAccentClassifier.__prepare_data(accents_dict, l)
            cv = ShuffleSplit(2, test_size=0.2, random_state=10)
            print("Cross-validating classifier with " + str(l) + " syllables...")
            cv_scores = cross_val_score(self.classifiers[l], train_data, answers, cv=cv, scoring='accuracy')
            result.append(cv_scores.mean())
        return result

    def classify_accent(self, word: str) -> int:
        """
        Проставление ударения в слове на основе классификатора.

        :param word: слово, в котором надо поставить ударение.
        :return: позиция буквы, на которую падает ударение.
        """
        syllables = Phonetics.get_word_syllables(word)
        answer = self.classifiers[len(syllables)].predict(np.array(self.__generate_sample(syllables)).reshape(1, -1))
        syllable = syllables[answer[0]]
        return get_first_vowel_position(syllable.text) + syllable.begin

    @staticmethod
    def __build_accent_classifier(model_dir: str, accents_dict: AccentDict, syllables_count: int) -> None:
        """
        Построение классификатора для заданного количества слогов.

        :param model_dir: папка, куда сохранить дамп.
        :param accents_dict: словарь ударений.
        :param syllables_count: количество слогов в обучающих примерах.
        """
        train_data, answers = MLAccentClassifier.__prepare_data(accents_dict, syllables_count)
        clf = DecisionTreeClassifier()
        clf.fit(train_data, answers)
        joblib.dump(clf, os.path.join(model_dir, MLAccentClassifier.clf_filename.format(syllables_count)))
        print("Built accent classifier for {syllables_count} syllables".format(syllables_count=syllables_count))

    @staticmethod
    def __prepare_data(accents_dict: AccentDict, syllables_count: int) -> Tuple[List[List[int]], List[int]]:
        """
        Подготовка данных для машинного обучения на оснвое словаря.

        :param accents_dict: словарь ударений.
        :param syllables_count: количество слогов, которое определяет классификатор.
        :return: X и y.
        """
        answers = []
        train_data = []
        for key, accents in accents_dict.get_all():
            syllables = Phonetics.get_word_syllables(key)
            if len(syllables) != syllables_count:
                continue
            for syllable in syllables:
                if "ё" in key:
                    continue
                primary_accents = [i[0] for i in accents if i[1] == AccentDict.AccentType.PRIMARY]
                if len(primary_accents) != 1:
                    continue
                for accent in primary_accents:
                    if syllable.begin <= accent < syllable.end:
                        syllable.accent = accent
                        answers.append(syllable.number)
                        train_data.append(MLAccentClassifier.__generate_sample(syllables))
        return train_data, answers

    @staticmethod
    def __generate_sample(syllables: List[Syllable]) -> List[int]:
        """
        По данному списку слогов формируется пример для машинного обучения.

        :param syllables: список слогов.
        :return: признаки.
        """
        l = len(syllables)
        features = []
        for i in range(0, l):
            text = syllables[i].text
            for j, ch in enumerate(text):
                if ch in VOWELS:
                    features.append(j)
            features.append(len(text))
            for ch1 in CYRRILIC_LOWER_CONSONANTS + CYRRILIC_LOWER_VOWELS:
                features.append(sum([ch1 == ch2 for ch2 in text]))
        return features



# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Тесты для разметки.

import unittest

from rupo.util.data import MARKUP_EXAMPLE
from rupo.main.markup import Markup


class TestMarkup(unittest.TestCase):
    def test_from_to(self):
        clean_markup = Markup()
        self.assertEqual(MARKUP_EXAMPLE, clean_markup.from_xml(MARKUP_EXAMPLE.to_xml()))
        clean_markup = Markup()
        self.assertEqual(MARKUP_EXAMPLE, clean_markup.from_json(MARKUP_EXAMPLE.to_json()))




# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Считыватель файлов разных расширений.

import os
import xml.etree.ElementTree as etree
import json
from enum import Enum
from typing import Iterator

from rupo.main.markup import Markup
from rupo.main.phonetics import Phonetics
from rupo.accents.dict import AccentDict
from rupo.accents.classifier import MLAccentClassifier
from rupo.metre.metre_classifier import MetreClassifier


RAW_SEPARATOR = "\n\n\n"


class FileTypeEnum(Enum):
    """
    Тип файла.
    """
    RAW = ".txt"
    XML = ".xml"
    JSON = ".json"
    STIHI = ""


class Reader(object):
    """
    Считывание из файлов.
    """
    @staticmethod
    def read_markups(path: str, source_type: FileTypeEnum, is_processed: bool,
                     accents_dict: AccentDict=None, accents_classifier: MLAccentClassifier=None) -> Iterator[Markup]:
        """
        Считывание разметок (включая разметку по сырым текстам).

        :param path: путь к файлу/папке.
        :param source_type: тип файлов.
        :param is_processed: уже размеченные тексты?
        :param accents_dict: словарь ударений (для неразмеченных текстов).
        :param accents_classifier: классификатор ударений (для неразмеченных текстов).
        """
        paths = Reader.__get_paths(path, source_type.value)
        for filename in paths:
            with open(filename, "r", encoding="utf-8") as file:
                if is_processed:
                    if source_type == FileTypeEnum.XML:
                        for elem in Reader.__xml_iter(file, 'markup'):
                            markup = Markup()
                            markup.from_xml(etree.tostring(elem, encoding='utf-8', method='xml'))
                            yield markup
                    elif source_type == FileTypeEnum.JSON:
                        j = json.load(file)
                        for item in j['items']:
                            markup = Markup()
                            markup.from_dict(item)
                            yield markup
                    elif source_type == FileTypeEnum.RAW:
                        raise NotImplementedError("Пока не реализовано.")
                else:
                    assert accents_dict is not None
                    assert accents_classifier is not None
                    for text in Reader.read_texts(filename, source_type):
                        yield Reader.__markup_text(text, accents_dict, accents_classifier)

    @staticmethod
    def read_texts(path: str, source_type: FileTypeEnum) -> Iterator[str]:
        """
        Считывание текстов.

        :param path: путь к файлу/папке.
        :param source_type: тип файлов.
        """
        paths = Reader.__get_paths(path, source_type.value)
        for filename in paths:
            with open(filename, "r", encoding="utf-8") as file:
                if source_type == FileTypeEnum.XML:
                    for elem in Reader.__xml_iter(file, 'item'):
                        yield elem.find(".//text").text
                elif source_type == FileTypeEnum.JSON:
                    # TODO: ленивый парсинг
                    j = json.load(file)
                    for item in j['items']:
                        yield item['text']
                elif source_type == FileTypeEnum.RAW:
                    text = file.read()
                    for t in text.split(RAW_SEPARATOR):
                        yield t
                elif source_type == FileTypeEnum.STIHI:
                    text = ""
                    is_text = False
                    for line in file:
                        if "<div" in line:
                            is_text = True
                        elif "</div>" in line:
                            is_text = False
                            yield text
                        elif is_text:
                            text += line + "\n"

    @staticmethod
    def __get_paths(path: str, ext: str) -> Iterator[str]:
        """
        Получение всех файлов заданного типа по заданному пути.

        :param path: путь к файлу/папке.
        :param ext: требуемое расширение.
        """
        if os.path.isfile(path):
            if ext == os.path.splitext(path)[1]:
                yield path
        else:
            for root, folders, files in os.walk(path):
                for file in files:
                    if ext == os.path.splitext(file)[1]:
                        yield os.path.join(root, file)
                for folder in folders:
                    return Reader.__get_paths(folder, ext)

    @staticmethod
    def __markup_text(text: str, accents_dict: AccentDict=None,
                      accents_classifier: MLAccentClassifier=None) -> Markup:
        """
        Разметка текста.

        :param text: текст.
        :param accents_dict: словарь ударений.
        :param accents_classifier: классификатор ударений.
        :return: разметка.
        """
        markup = Phonetics.process_text(text, accents_dict)
        markup = MetreClassifier.improve_markup(markup, accents_classifier)[0]
        return markup

    @staticmethod
    def __xml_iter(file, tag):
        """
        :param file: xml файл.
        :param tag: заданный тег.
        :return: все элементы с заданными тегами в xml.
        """
        return (elem for event, elem in etree.iterparse(file, events=['end']) if event == 'end' and elem.tag == tag)



# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Тесты для API библиотеки.

import unittest

from rupo.main.markup import Markup
from rupo.api import get_accent, get_word_syllables, count_syllables, get_markup, get_improved_markup, \
    is_rhyme, classify_metre


class TestApi(unittest.TestCase):
    def test_accent(self):
        self.assertEqual(get_accent("корова"), 3)

    def test_get_word_syllables(self):
        self.assertEqual(get_word_syllables("корова"), ["ко", "ро", "ва"])

    def test_count_syllables(self):
        self.assertEqual(count_syllables("корова"), 3)

    def test_is_rhyme(self):
        self.assertTrue(is_rhyme("корова", "здорова"))

    def test_get_markup(self):
        self.assertIsInstance(get_markup("корова"), Markup)

    def test_get_improved_markup(self):
        self.assertIsInstance(get_improved_markup("корова"), Markup)

    def test_classify_metre(self):
        text = "Горит восток зарёю новой.\n" \
               "Уж на равнине, по холмам\n" \
               "Грохочут пушки. Дым багровый\n" \
               "Кругами всходит к небесам."
        self.assertEqual(classify_metre(text), "iambos")



# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Класс для удобной работы со словарём ударений.

import datrie
import os
from typing import List, Tuple
from enum import Enum

from rupo.util.preprocess import CYRRILIC_LOWER_VOWELS, CYRRILIC_LOWER_CONSONANTS
from rupo.settings import DICT_TXT_PATH
from rupo.settings import DICT_TRIE_PATH


class AccentDict:
    """
    Класс данных, для сериализации словаря как dict'а и быстрой загрузки в память.
    """

    class AccentType(Enum):
        ANY = -1
        PRIMARY = 0
        SECONDARY = 1

    AccentType = AccentType

    def __init__(self) -> None:
        self.data = datrie.Trie(CYRRILIC_LOWER_VOWELS+CYRRILIC_LOWER_CONSONANTS+"-")
        src_filename = DICT_TXT_PATH
        dst_filename = DICT_TRIE_PATH
        if not os.path.isfile(src_filename):
            print(src_filename)
            raise FileNotFoundError("Не найден файл словаря.")
        if os.path.isfile(dst_filename):
            self.data = datrie.Trie.load(dst_filename)
        else:
            self.create(src_filename, dst_filename)
        print("Accent dict loaded")

    def create(self, src_filename: str, dst_filename: str) -> None:
        """
        Загрузка словаря из файла. Если уже есть его сериализация в .trie файле, берём из него.

        :param src_filename: имя файла с оригинальным словарём.
        :param dst_filename: имя файла, в который будет сохранён дамп.
        """
        with open(src_filename, 'r', encoding='utf-8') as f:
            lines = f.readlines()
            for line in lines:
                for word in line.split("#")[1].split(","):
                    word = word.strip()
                    pos = -1
                    accents = []
                    clean_word = ""
                    for i in range(len(word)):
                        if word[i] == "'" or word[i] == "`":
                            if word[i] == "`":
                                accents.append((pos, AccentDict.AccentType.SECONDARY))
                            else:
                                accents.append((pos, AccentDict.AccentType.PRIMARY))
                            continue
                        clean_word += word[i]
                        pos += 1
                        if word[i] == "ё":
                            accents.append((pos, AccentDict.AccentType.PRIMARY))
                    self.__update(clean_word, accents)
        self.data.save(dst_filename)

    def save(self, dst_filename: str) -> None:
        """
        Сохранение дампа.

        :param dst_filename: имя файла, в который сохраняем дамп словаря.
        """
        self.data.save(dst_filename)

    def get_accents(self, word: str, accent_type: AccentType=AccentType.ANY) -> List[int]:
        """
        Обёртка над data.get().

        :param word: слово, которое мы хотим посмотреть в словаре.
        :param accent_type: тип ударения.
        :return forms: массив всех ударений.
        """
        if word in self.data:
            if accent_type == AccentDict.AccentType.ANY:
                return [i[0] for i in self.data[word]]
            else:
                return [i[0] for i in self.data[word] if i[1] == accent_type]
        return []

    def get_all(self) -> List[Tuple[str, List[Tuple[int, AccentType]]]]:
        """
        :return items: все ключи и ударения словаря.
        """
        return self.data.items()

    def __update(self, word: str, accent_pairs: List[Tuple[int, AccentType]]) -> None:
        """
        Обновление словаря.

        :param word: слово.
        :param accent_pairs: набор ударений.
        """
        if word not in self.data:
            self.data[word] = set(accent_pairs)
        else:
            self.data[word].update(accent_pairs)



# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Модуль для описания разметки по ударениям и слогам.

import json
from typing import List
import xml.etree.ElementTree as etree

from dicttoxml import dicttoxml

from rupo.util.preprocess import get_first_vowel_position
from rupo.util.mixins import CommonMixin


class Annotation(CommonMixin):
    """
    Класс аннотации.
    Содержит начальную и конечную позицию в тексте, а также текст аннотации .
    """
    def __init__(self, begin: int, end: int, text: str) -> None:
        self.begin = begin
        self.end = end
        self.text = text


class Syllable(Annotation):
    """
    Разметка слога. Включает в себя аннотацию и номер слога, а также ударение.
    Если ударение падает не на этот слог, -1.
    """
    def __init__(self, begin: int, end: int, number: int, text: str, accent: int=-1) -> None:
        super(Syllable, self).__init__(begin, end, text)
        self.number = number
        self.accent = accent

    def vowel(self) -> int:
        """
        :return: позиция гласной буквы этого слога в слове (с 0).
        """
        return get_first_vowel_position(self.text) + self.begin

    def from_dict(self, d: dict) -> 'Syllable':
        self.__dict__.update(d)
        return self


class Word(Annotation):
    """
    Разметка слова. Включает в себя аннотацию слова и его слоги.
    """
    def __init__(self, begin: int, end: int, text: str, syllables: List[Syllable]) -> None:
        super(Word, self).__init__(begin, end, text)
        self.syllables = syllables

    def count_accents(self) -> int:
        """
        :return: количество ударений в слове.
        """
        return sum(syllable.accent != -1 for syllable in self.syllables)

    def accent(self) -> int:
        """
        :return: последнее ударение в слове, если нет, то -1.
        """
        accent = -1
        for syllable in self.syllables:
            if syllable.accent != -1:
                accent = syllable.accent
        return accent

    def get_accented_syllables_numbers(self) -> List[int]:
        """
        :return: номера слогов, на которые падают ударения.
        """
        return [syllable.number for syllable in self.syllables if syllable.accent != -1]

    def set_accents(self, accents: List[int]) -> None:
        """
        Задать ударения, все остальные убираются.

        :param accents: позиции ударения в слове.
        """
        for syllable in self.syllables:
            if syllable.vowel() in accents:
                syllable.accent = syllable.vowel()
            else:
                syllable.accent = -1

    def get_short(self) -> str:
        """
        :return: слово в форме "текст"+"последнее ударение".
        """
        return self.text.lower() + str(self.accent())

    def from_dict(self, d: dict) -> 'Word':
        self.__dict__.update(d)
        syllables = d["syllables"]  # type: List[dict]
        self.syllables = [Syllable(0, 0, 0, "").from_dict(syllables[i]) for i in range(len(syllables))]
        return self

    def __hash__(self) -> int:
        """
        :return: хеш разметки.
        """
        return hash(self.get_short())


class Line(Annotation):
    """
    Разметка строки. Включает в себя аннотацию строки и её слова.
    """
    def __init__(self, begin: int, end: int, text: str, words: List[Word]) -> None:
        super(Line, self).__init__(begin, end, text)
        self.words = words

    def from_dict(self, d) -> 'Line':
        self.__dict__.update(d)
        words = d["words"]  # type: List[dict]
        self.words = [Word(0, 0, "", []).from_dict(words[i]) for i in range(len(words))]
        return self


class Markup(CommonMixin):
    """
    Класс данных для разметки в целом с экспортом/импортом в XML и JSON.
    """
    def __init__(self, text: str=None, lines: List[Line]=None) -> None:
        # TODO: При изменении структуры разметки менять десериализацию.
        self.text = text
        self.lines = lines
        self.version = 2

    def to_json(self) -> str:
        return json.dumps(self.to_dict(), ensure_ascii=False)

    def from_json(self, st) -> 'Markup':
        d = json.loads(st)
        return self.from_dict(d)

    def from_dict(self, d) -> 'Markup':
        self.__dict__.update(d)
        lines = d["lines"]  # type: List[dict]
        self.lines = [Line(0, 0, "", []).from_dict(lines[i]) for i in range(len(lines))]
        return self

    def to_xml(self) -> str:
        """
        Экспорт в XML.

        :return self: строка в формате XML
        """
        return dicttoxml(self.to_dict(), custom_root='markup', attr_type=False).decode('utf-8').replace("\n", "\\n")

    def from_xml(self, xml: str) -> 'Markup':
        """
        Импорт из XML.

        :param xml: XML-разметка
        :return self: получившийся объект Markup
        """
        root = etree.fromstring(xml)
        if root.find("version") is None or int(root.find("version").text) != self.version:
            raise TypeError("Другая версия разметки")
        lines_node = root.find("lines")
        lines = []
        for line_node in lines_node.findall("item"):
            words_node = line_node.find("words")
            words = []
            for word_node in words_node.findall("item"):
                syllables_node = word_node.find("syllables")
                syllables = []
                for syllable_node in syllables_node.findall("item"):
                    syllables.append(Syllable(int(syllable_node.find("begin").text),
                                              int(syllable_node.find("end").text),
                                              int(syllable_node.find("number").text),
                                              syllable_node.find("text").text,
                                              int(syllable_node.find("accent").text)))
                words.append(Word(int(word_node.find("begin").text), int(word_node.find("end").text),
                                  word_node.find("text").text, syllables))
            lines.append(Line(int(line_node.find("begin").text), int(line_node.find("end").text),
                              line_node.find("text").text, words))
        self.text = root.find("text").text.replace("\\n", "\n")
        self.lines = lines
        return self



# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Набор внешних методов для работы с библиотекой.

from typing import List

from rupo.main.phonetics import Phonetics
from rupo.main.markup import Markup
from rupo.accents.dict import AccentDict
from rupo.accents.classifier import MLAccentClassifier
from rupo.metre.metre_classifier import MetreClassifier
from rupo.files.reader import FileTypeEnum, Reader
from rupo.files.writer import Writer
from rupo.rhymes.rhymes import Rhymes
from rupo.generate.markov import MarkovModelContainer
from rupo.generate.generator import Generator


class Global:
    """
    Глобальные ресурсы.
    """
    accent_dict = None
    accent_classifier = None
    markov = None
    generator = None

    @classmethod
    def get_dict(cls):
        if cls.accent_dict is None:
            cls.accent_dict = AccentDict()
        return cls.accent_dict

    @classmethod
    def get_classifier(cls):
        if cls.accent_classifier is None:
            cls.accent_classifier = MLAccentClassifier(cls.get_dict())
        return cls.accent_classifier

    @classmethod
    def get_markov(cls, markup_path, dump_path):
        if cls.markov is None:
            cls.markov = MarkovModelContainer(dump_path, markup_path)
        return cls.markov

    @classmethod
    def get_generator(cls, markup_path, dump_path):
        if cls.generator is None:
            cls.generator = Generator(cls.get_markov(markup_path, dump_path),
                                      cls.get_markov(markup_path, dump_path).vocabulary)
        return cls.generator


def get_accent(word: str) -> int:
    """
    :param word: слово.
    :return: ударение слова.
    """
    return Phonetics.get_improved_word_accent(word, Global.get_dict(), Global.get_classifier())


def get_word_syllables(word: str) -> List[str]:
    """
    :param word: слово.
    :return: его слоги.
    """
    return [syllable.text for syllable in Phonetics.get_word_syllables(word)]


def count_syllables(word: str) -> int:
    """
    :param word: слово.
    :return: количество слогов в нём.
    """
    return len(Phonetics.get_word_syllables(word))


def get_markup(text: str) -> Markup:
    """
    :param text: текст.
    :return: его разметка по словарю.
    """
    return Phonetics.process_text(text, Global.get_dict())


def get_improved_markup(text: str) -> Markup:
    """
    :param text: текст.
    :return: его разметка по словарю, классификатору метру и  ML классификатору.
    """
    markup = Phonetics.process_text(text, Global.get_dict())
    return MetreClassifier.improve_markup(markup, Global.get_classifier())[0]


def classify_metre(text: str) -> str:
    """
    :param text: текст.
    :return: его метр.
    """
    return MetreClassifier.classify_metre(Phonetics.process_text(text, Global.get_dict())).metre


def generate_markups(input_path: str, input_type: FileTypeEnum, output_path: str, output_type: FileTypeEnum) -> None:
    """
    Генерация разметок по текстам.

    :param input_path: путь к папке/файлу с текстом.
    :param input_type: тип файлов с текстов.
    :param output_path: путь к файлу с итоговыми разметками.
    :param output_type: тип итогового файла.
    """
    markups = Reader.read_markups(input_path, input_type, False, Global.get_dict(), Global.get_classifier())
    writer = Writer(output_type, output_path)
    writer.open()
    for markup in markups:
        writer.write_markup(markup)
    writer.close()


def is_rhyme(word1: str, word2: str) -> bool:
    """
    :param word1: первое слово.
    :param word2: второе слово.
    :return: рифмуются ли слова.
    """
    markup_word1 = get_markup(word1).lines[0].words[0]
    markup_word1.set_accents([get_accent(word1)])
    markup_word2 = get_markup(word2).lines[0].words[0]
    markup_word2.set_accents([get_accent(word2)])
    return Rhymes.is_rhyme(markup_word1, markup_word2)


def generate_poem(markup_path, dump_path, metre_schema: str="-+",
                  rhyme_pattern: str="abab", n_syllables: int=8) -> str:
    """
    Сгенерировать стих по данным из разметок.

    :param markup_path: путь к разметкам.
    :param dump_path: путь, куда сохранять модель.
    :param metre_schema: схема метра.
    :param rhyme_pattern: схема рифм.
    :param n_syllables: количество слогов в строке.
    :return: стих.
    """
    generator = Global.get_generator(dump_path, markup_path)
    return generator.generate_poem(metre_schema, rhyme_pattern, n_syllables)


def generate_poem_by_line(markup_path, dump_path, line, rhyme_pattern="abab") -> str:
    """
    Сгенерировать стих по первой строчке.

    :param markup_path: путь к разметкам.
    :param dump_path: путь, куда сохранять модель.
    :param line: первая строчка
    :param rhyme_pattern: схема рифм.
    :return: стих.
    """
    generator = Global.get_generator(dump_path, markup_path)
    return generator.generate_poem_by_line(line, rhyme_pattern, Global.get_dict(), Global.get_classifier())



from setuptools import find_packages, setup
from setuptools.command.develop import develop
from setuptools.command.install import install
from rupo.accents.dict import AccentDict
from rupo.accents.classifier import MLAccentClassifier


class PostDevelopCommand(develop):
    def run(self):
        d = AccentDict()
        MLAccentClassifier(d)
        develop.run(self)


class PostInstallCommand(install):
    def run(self):
        d = AccentDict()
        MLAccentClassifier(d)
        install.run(self)


setup(
    name='rupo',
    packages=find_packages(),
    version='0.1',
    description='RuPo: library for russian poetry analysis and generation',
    author='Ilya Gusev',
    author_email='phoenixilya@gmail.com',
    url='https://github.com/IlyaGusev/rupo',
    download_url='https://github.com/IlyaGusev/rupo/archive/0.1.tar.gz',
    keywords=['poetry', 'nlp', 'russian'],
    classifiers=[],
    package_data={
        'rupo': ['data/dict/*.txt', 'data/dict/*.trie', 'data/classifier/*.pickle'],
    },
    install_requires=[
        'dicttoxml==1.7.4',
        'datrie==0.7.1',
        'numpy==1.11.3',
        'scipy==0.18.1',
        'scikit-learn==0.18.1',
        'jsonpickle==0.9.4'
    ],
    cmdclass={
        'develop': PostDevelopCommand,
        'install': PostInstallCommand,
    },
)


# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Модуль разбивки на слоги, проставления ударений и получения начальной разметки.

from typing import List

from rupo.accents.dict import AccentDict
from rupo.main.markup import Syllable, Word, Markup, Line
from rupo.util.preprocess import count_vowels, get_first_vowel_position, \
    VOWELS, CLOSED_SYLLABLE_CHARS


class Phonetics:
    """
    Класс-механизм для фонетического анализа слов.
    """
    @staticmethod
    def get_word_syllables(word: str) -> List[Syllable]:
        """
        Разделение слова на слоги.

        :param word: слово для разбивки на слоги.
        :return syllables: массив слогов слова.
        """
        syllables = []
        begin = 0
        number = 0
        for i in range(len(word)):
            if word[i] not in VOWELS:
                continue
            if i+1 < len(word)-1 and word[i+1] in CLOSED_SYLLABLE_CHARS:
                if i+2 < len(word)-1 and word[i+2] in "ьЬ":
                    # Если после сонорного согласного идёт мягкий знак, заканчиваем на нём. ("бань-ка")
                    end = i+3
                elif i+2 < len(word)-1 and word[i+2] not in VOWELS and \
                        (word[i+2] not in CLOSED_SYLLABLE_CHARS or word[i+1] == "й"):
                    # Если после сонорного согласного не идёт гласная или другой сонорный согласный,
                    # слог закрывается на этом согласном. ("май-ка")
                    end = i+2
                else:
                    # Несмотря на наличие закрывающего согласного, заканчиваем на гласной.
                    # ("со-ло", "да-нный", "пол-ный")
                    end = i+1
            else:
                # Если после гласной идёт не закрывающая согласная, заканчиваем на гласной. ("ко-гда")
                end = i+1
            syllables.append(Syllable(begin, end, number, word[begin:end]))
            number += 1
            begin = end
        if get_first_vowel_position(word) != -1:
            # Добиваем последний слог до конца слова.
            syllables[-1] = Syllable(syllables[-1].begin, len(word), syllables[-1].number,
                                     word[syllables[-1].begin:len(word)])
        return syllables

    @staticmethod
    def get_word_accents(word: str, accents_dict: AccentDict) -> List[int]:
        """
        Определение ударения в слове по словарю. Возможно несколько вариантов ударения.

        :param word: слово для простановки ударений.
        :param accents_dict: экземпляр обёртки для словаря ударений.
        :return accents: позиции букв, на которые падает ударение.
        """
        accents = []
        if count_vowels(word) == 0:
            # Если гласных нет, то и ударений нет.
            pass
        elif count_vowels(word) == 1:
            # Если одна гласная, то на неё и падает ударение.
            accents.append(get_first_vowel_position(word))
        elif word.find("ё") != -1:
            # Если есть буква "ё", то только на неё может падать ударение.
            accents.append(word.find("ё"))
        else:
            # Проверяем словарь на наличие форм с ударениями.
            accents = accents_dict.get_accents(word)
            if 'е' not in word:
                return accents
            # Находим все возможные варинаты преобразований 'е' в 'ё'.
            positions = [i for i in range(len(word)) if word[i] == 'е']
            beam = [word[:positions[0]]]
            for i in range(len(positions)):
                new_beam = []
                for prefix in beam:
                    n = positions[i+1] if i+1 < len(positions) else len(word)
                    new_beam.append(prefix + 'ё' + word[positions[i]+1:n])
                    new_beam.append(prefix + 'е' + word[positions[i]+1:n])
                    beam = new_beam
            # И проверяем их по словарю.
            for permutation in beam:
                if len(accents_dict.get_accents(permutation)) != 0:
                    yo_pos = permutation.find("ё")
                    if yo_pos != -1:
                        accents.append(yo_pos)
        return accents

    @staticmethod
    def process_text(text: str, accents_dict: AccentDict) -> Markup:
        """
        Получение начального варианта разметки по слогам и ударениям.

        :param text: текст для разметки
        :param accents_dict: экземпляр обёртки для словаря ударений
        :return markup: разметка по слогам и ударениям
        """
        begin_word = -1
        begin_line = 0
        lines = []
        words = []
        # TODO: Нормальная токенизация.
        for i in range(len(text)):
            valid_word_symbol = text[i].isalpha() and i != len(text) - 1
            if valid_word_symbol and begin_word == -1:
                begin_word = i
            if not valid_word_symbol and begin_word != -1:
                # Каждое слово разбиваем на слоги.
                word = Word(begin_word, i, text[begin_word:i], Phonetics.get_word_syllables(text[begin_word:i]))
                # Проставляем ударения.
                accents = Phonetics.get_word_accents(word.text.lower(), accents_dict)
                # Сопоставляем ударения слогам.
                word.set_accents(accents)
                words.append(word)
                begin_word = -1
            if text[i] == "\n":
                # Разбиваем по строкам.
                lines.append(Line(begin_line, i+1, text[begin_line:i], words))
                words = []
                begin_line = i+1
        if begin_line != len(text):
            lines.append(Line(begin_line, len(text), text[begin_line:len(text)], words))
        return Markup(text, lines)

    @staticmethod
    def get_improved_word_accent(word: str, accent_dict: AccentDict, accent_classifier) -> int:
        """
        Получение ударения с учётом классификатора.

        :param word: слово.
        :param accent_dict: словарь ударений.
        :param accent_classifier: классификатор ударений.
        :return: индекс ударения.
        """
        dict_accents = Phonetics.get_word_accents(word, accent_dict)
        if len(dict_accents) == 1:
            return dict_accents[0]
        elif len(dict_accents) == 0:
            clf_accent = accent_classifier.classify_accent(word)
            return clf_accent
        else:
            clf_accent = accent_classifier.classify_accent(word)
            intersection = list(set(dict_accents).intersection({clf_accent}))
            if len(intersection) != 0:
                return intersection[0]
            else:
                return dict_accents[0]
